use std::{collections::HashMap, fs, io, path::PathBuf, process::exit};

use async_trait::async_trait;
use dialoguer::{Input, Select};
use encoding::{all::ISO_8859_1, EncoderTrap, Encoding};
use serde::Deserialize;
use tokio::{fs::File, io::AsyncReadExt};

use crate::types::{message::{Message, Photo}, participant::Participant, project::ProjectRaw, user::User};

use super::service::Service;

fn decode_str(input: &String) -> String {
    let vec = ISO_8859_1.encode(input.as_str(), EncoderTrap::Strict).unwrap();
    String::from_utf8(vec).unwrap()
}

#[derive(Deserialize, Debug)]
pub struct InstagramMessage {
    sender_name: String,
    timestamp_ms: usize,
    content: Option<String>,
    photos: Option<Vec<InstagramPhoto>>
}

impl InstagramMessage {
    // TODO: Take into consideration the user_map
    pub fn conv(&self, media_index: &mut usize, project_path: PathBuf, user_map: &HashMap<Participant, User>) -> Message {
        let text_content = if let Some(content) = &self.content {
            Some(decode_str(content))
        } else {
            None
        };

        let photos: Vec<Photo> = if let Some(photos_list) = &self.photos {
            photos_list.iter().map(|photo| photo.conv(&text_content, media_index, project_path.clone())).collect()
        } else {
            vec![]
        };

        let participant = Participant {
            name: decode_str(&self.sender_name)
        };

        Message {
            sender_name: user_map.get(&participant).unwrap().name.clone(),
            text_content,
            timestamp: self.timestamp_ms,
            photos
        }
    }
}

#[derive(Deserialize, Debug)]
struct InstagramPhoto {
    uri: String
}

impl InstagramPhoto {
    pub fn conv(&self, text_content: &Option<String>, media_index: &mut usize, mut original_path: PathBuf) -> Photo {
        *media_index += 1;
        let img_path = (*media_index - 1).to_string();

        original_path.pop();

        original_path.push(self.uri.clone());

        Photo {
            text_content: text_content.clone(),
            img_path,
            original_img_path: original_path.display().to_string()
        }
    }
}

#[derive(Deserialize, Debug)]
pub struct InstagramParticipant {
    pub name: String,
}

#[derive(Deserialize, Debug)]
struct InstagramDataJson {
    messages: Vec<InstagramMessage>,
    participants: Vec<InstagramParticipant>
}

pub struct InstagramData {
    pub project_path: String,
    pub participants: Vec<InstagramParticipant>,
    pub messages: Vec<InstagramMessage>,
}

pub struct Instagram {
    pub data: Option<InstagramData>,
}

#[async_trait]
impl Service for Instagram {
    fn get_name(&self) -> &'static str {
        "Instagram"
    }

    fn setup(&self) -> Box<dyn Service> {
        let export_path: String = Input::new()
            .with_prompt("Enter path for your instagram export. (your_instagram_activity folder)")
            .interact()
            .unwrap();

        let mut path = PathBuf::from(export_path);

        path.push(PathBuf::from("messages/inbox"));

        let options: Vec<String> = fs::read_dir(&path).unwrap_or_else(|_| {
            eprintln!("Folder not found!");
            exit(1);
        })
            .filter_map(|entry| entry.ok())
            .filter(|entry| entry.path().is_dir())
            .map(|entry| { String::from(entry.file_name().to_str().unwrap()) }).collect();

        let option = Select::new()
            .with_prompt("Choose a chat")
            .items(&options)
            .default(0)
            .interact()
            .unwrap();

        let selected = options.get(option).unwrap().clone();

        path.push(selected.clone());

        Box::new(Instagram {
            data: Some(InstagramData {
                project_path: path.display().to_string(),
                participants: vec![],
                messages: vec![],
            }),
        })
    }

    async fn load(&mut self) -> io::Result<()> {
        let data = self.data.as_mut().unwrap();
        let path = PathBuf::from(&data.project_path);

        let paths: Vec<PathBuf> = fs::read_dir(&path).unwrap_or_else(|_| {
            eprintln!("Folder not found!");
            exit(1);
        })
            .filter_map(|entry| entry.ok())
            .filter(|entry| entry.path().is_file())
            .map(|entry| { entry.path() }).collect();

        for (index, path) in paths.iter().enumerate() {
            if !path.exists() {
                return Err(io::Error::new(
                    io::ErrorKind::NotFound,
                    format!(
                        "File does not exist: {}",
                        path.to_str().unwrap(),
                    ),
                ));
            }

            let mut file = File::open(path).await?;
            let mut contents = String::new();
            file.read_to_string(&mut contents).await?;

            let jsondata: InstagramDataJson = serde_json::from_str(&contents)
                .map_err(|e| io::Error::new(io::ErrorKind::InvalidData, e))?;

            if index == 0{
                // Load participants
                for participant in jsondata.participants {
                    data.participants.push(participant);
                }
            }

            for message in jsondata.messages {
                data.messages.push(message);
            }
        }
        
        Ok(())
    }

    fn get_participants(&self) -> Vec<Participant> {
        let data = self.data.as_ref().unwrap();

        data.participants.iter().map(|participant| {
            Participant {
                name: decode_str(&participant.name)
            }
        }).collect()
    }

    fn merge_messages(&self, raw_project: &mut ProjectRaw, user_map: &HashMap<Participant, User>) {
        let data = self.data.as_ref().unwrap();
        let path = PathBuf::from(data.project_path.clone());

        for message in &data.messages {
            let mut media_index_clone = raw_project.project.media_index.clone();
            let msg = message.conv(&mut media_index_clone, path.clone(), user_map);

            let moved = raw_project.project.push_msg(&msg);

            if moved {
                // Move the file to the media folder
                

                // Update the highest media index of the project if the message was merged
                raw_project.project.media_index = media_index_clone;
            }

            println!("{:?}", msg);
        }
    }
}
